#include "../global.h"

#include <time.h>
#include <logger/Log.h>

#include <db/DoubleDatabase.h>
#include "../translate/TranslationRule.h"
#include "../translate/TranslationTable.h"
#include "../translate/CandidateRule.h"

#include "ExactSearch.h"
#include "SelectSearch.h"
#include "GreedySearch.h"
#include "KrimpSearch.h"

#include "RemodelSearch.h"

#include <sys/timeb.h>

RemodelSearch::RemodelSearch() {
	mDDB = NULL;
	mProgress = 0;
	mProgressFile = NULL;
}
RemodelSearch::~RemodelSearch() {

}
void RemodelSearch::Init(DoubleDatabase *ddb) {
	mDDB = ddb;
	mProgress = 0;
}
RemodelSearch* RemodelSearch::Create(const string &searchName, Config *config) {
	if(searchName.compare("select") == 0)
		return new SelectSearch(config);
	if (searchName.compare("greedy") == 0)
		return new GreedySearch(config);
	if (searchName.compare("exact") == 0)
		return new ExactSearch(config);
	if(searchName.compare("krimp") == 0)
		return new KrimpSearch(config);
	THROW("Unknown RemodelSearch type: " + searchName);
}
void RemodelSearch::ProgressToScreen(const uint64 curCandidate, const uint64 total) {
	if(Bass::GetOutputLevel() > 0) {
		uint16 p = (uint16) (100 * curCandidate / total);

		if(mProgress < p) {
			mProgress = p;

			// Percentage uitrekenen
			int d = p % 4;
			char c = '*';
			switch(d) {
			case 0 : c = '-'; break;
			case 1 : c = '\\'; break;
			case 2 : c = '|'; break;
			default : c = '/';
			}
#if defined (_MSC_VER) && defined (_WINDOWS)
			printf(" %c Progress:\t\t%d%% (%I64d)   \t\t\r", c, p, curCandidate);
#elif defined (__GNUC__) && defined (__unix__)
			printf(" %c Progress:\t\t%d%% (%lu)   \t\t\r", c, p, curCandidate);
#endif
		}
	}
}
void RemodelSearch::OpenProgressFile() {
	if(mProgressFile!=NULL)
		THROW("Progress file already opened?");
	string filename = Bass::GetWorkingDir() + "report-" + Logger::GetTimestamp() + ".csv";

	// Open file for writing
	if(fopen_s(&mProgressFile, filename.c_str(), "w") != 0) {
		printf("Cannot open progress report file for writing!\nFilename: %s\n", filename.c_str());
		if(fopen_s(&mProgressFile, filename.c_str(), "w") != 0) {
			printf("mProgressFile try failed as well..\n");
			THROW("Could not write progress file, aborting.");
		}
	}
	// Header
	fprintf_s(mProgressFile, "action;candId;#rules;L(L | TT);L(R | TT);L(TT);L(total);#uncoveredL;#errorsL;#uncoveredR;#errorsR;rule;stepEls(ms);writeTmstmp\n");
	fflush(mProgressFile);
}


void RemodelSearch::ProgressToFileAddRule(TranslationRule *rule, const uint64 candId, TranslationStats &stats, int time_ms) {

#if defined (_MSC_VER) && defined (_WINDOWS)
  fprintf_s(mProgressFile, "+;%I64d;%u;%.0f;%.0f;%.0f;%.0f;%u;%u;%u;%u;%s;%u;%u\n", candId, stats.numRules, stats.encLeftSize, stats.encRightSize, stats.encTableSize, stats.encSize, stats.nbUncoveredL, stats.nbErrorsL, stats.nbUncoveredR, stats.nbErrorsR, rule->ToString().c_str(), time_ms, time(NULL));
#elif defined (__GNUC__) && defined (__unix__)
  fprintf_s(mProgressFile, "+;%lu;%u;%.0f;%.0f;%.0f;%.0f;%u;%u;%u;%u;%s;%u;%u\n", candId, stats.numRules, stats.encLeftSize, stats.encRightSize, stats.encTableSize, stats.encSize, stats.nbUncoveredL, stats.nbErrorsL, stats.nbUncoveredR, stats.nbErrorsR, rule->ToString().c_str(), time_ms, time(NULL));
#endif
  fflush(mProgressFile);
}

void RemodelSearch::ProgressToFileCandidate(CandidateRule *rule, const uint64 candId) {
#if defined (_MSC_VER) && defined (_WINDOWS)
  fprintf_s(mProgressFile, "c;%I64d;;;;;%.0f;;;;;%s <-> %s\n", candId, rule->GetGain(), rule->GetLeft()->ToString().c_str(), rule->GetRight()->ToString().c_str());
#elif defined (__GNUC__) && defined (__unix__)
	fprintf_s(mProgressFile, "c;%lu;;;;;%.0f;;;;;%s <-> %s\n", candId, rule->GetGain(), rule->GetLeft()->ToString().c_str(), rule->GetRight()->ToString().c_str());
#endif
	fflush(mProgressFile);
}


void RemodelSearch::CloseProgressFile() {
	fclose(mProgressFile);
	mProgressFile = NULL;
}

int RemodelSearch::getMilliCount(){
  timeb tb;
  ftime(&tb);
  int nCount = tb.millitm + (tb.time & 0xfffff) * 1000;
  return nCount;
}

int RemodelSearch::getMilliSpan(int nTimeStart){
  int nSpan = getMilliCount() - nTimeStart;
  if(nSpan < 0)
    nSpan += 0x100000 * 1000;
  return nSpan;
}

