#include "../global.h"

#if defined (_WINDOWS)
	#include <direct.h>
#endif
#include <time.h>

// qtils
#include <Config.h>
#include <FileUtils.h>
#include <StringUtils.h>
#include <TimeUtils.h>
#include <logger/Log.h>

// bass
#include <Bass.h>
#include <db/DoubleDatabase.h>
#include <isc/ItemSetCollection.h>
#include "../FicConf.h"

#include "../search/RemodelSearch.h"
#include "../translate/TranslationTable.h"

#include "TranslatorTH.h"

TranslatorTH::TranslatorTH(Config *conf) : TaskHandler(conf){

}

TranslatorTH::~TranslatorTH() {
	// not my Config*
}

void TranslatorTH::HandleTask() {
	string command = mConfig->Read<string>("command");

			if(command.compare("translator") == 0)		Translator();
	else	if(command.compare("ttencsize") == 0)		ComputeTranslationTableEncodedSize();

	else	throw string("TranslatorTH :: Unable to handle task `" + command + "`");
}

string TranslatorTH::BuildWorkingDir() {
	string command = mConfig->Read<string>("command");

	if(command.compare("translator") == 0)	{
		if(mConfig->KeyExists("iscName"))
			return mConfig->Read<string>("command") + "/" + mConfig->Read<string>("search") + "-" + mConfig->Read<string>("iscName") + "-" + Logger::GetTimestamp() + "/";
		else
			return mConfig->Read<string>("command") + "/" + mConfig->Read<string>("search") + "-" + mConfig->Read<string>("dbLeft") + "-" + mConfig->Read<string>("dbRight") + "-" + Logger::GetTimestamp() + "/";

	} else	if(command.compare("ttencsize") == 0) { 
		return mConfig->Read<string>("command") + "/" + mConfig->Read<string>("ttName") + "/";
	}
	THROW("Unknown task.");
}


/* ------------------------- Commands -------------------------- */

void TranslatorTH::Translator() {
	char temp[500]; string s;

	// Load the data
	Database *dbL = Database::RetrieveDatabase(mConfig->Read<string>("dbLeft"), Uint16ItemSetType);
	Database *dbR = Database::RetrieveDatabase(mConfig->Read<string>("dbRight"), Uint16ItemSetType);
	DoubleDatabase *ddb = new DoubleDatabase(dbL, dbR); // takes ownership of dbL and dbR
	
	// Not very nice, possibly loading the data twice, but this is an easy solution for now; should give significant speedups when possible..
	if(ddb->GetAlphabetSize() < 129) {
		delete ddb;
		dbL = Database::RetrieveDatabase(mConfig->Read<string>("dbLeft"), BM128ItemSetType);
		dbR = Database::RetrieveDatabase(mConfig->Read<string>("dbRight"), BM128ItemSetType);
		ddb = new DoubleDatabase(dbL, dbR);
	}

	// Write DDB to disk if desired
	//Database::WriteDatabase(ddb, "doubledata", Bass::GetWorkingDir());

	// -- Obtain TranslationTable through search --
	LOG_MSG("** Starting remodel search for TranslationTable");
	RemodelSearch *search = RemodelSearch::Create(mConfig->Read<string>("search", "select"), mConfig);
	search->Init(ddb);
	TranslationTable *tt = search->MineThatTable();
	LOG_MSG("Finished search.\t\t\t\t\n\n** The results");

	// Output results to screen
	sprintf_s(temp, 500, " # rules:\t\t%u", tt->GetNumRules()); s = temp; LOG_MSG(s);
	sprintf_s(temp, 500, " Compressed size:\t%.2f bits", tt->GetEncodedSize()); s = temp; LOG_MSG(s);
	double stdSize = ddb->GetLeftDatabase()->GetStdDbSize() + ddb->GetRightDatabase()->GetStdDbSize();
	sprintf_s(temp, 500, " Standard size:\t\t%.2f bits", stdSize); s = temp; LOG_MSG(s);
	sprintf_s(temp, 500, " Compression ratio:\t%.2f%%\n", 100.0 * (tt->GetEncodedSize() / stdSize)); s = temp; LOG_MSG(s);

	// Write TT to disk
	string filename = Bass::GetWorkingDir() + "rules-" + Logger::GetTimestamp() + ".tt";
	tt->WriteToDisk(filename);

	// Delete mined TT and search
	delete tt;
	delete search;

	// Load TT from disk and verify whether it gives the same results
	bool verify = false;
	if(verify) {
		tt = TranslationTable::LoadTranslationTable(filename, ddb);
		tt->Translate();
		sprintf_s(temp, 500, " # rules:\t\t%u", tt->GetNumRules()); s = temp; LOG_MSG(s);
		sprintf_s(temp, 500, " Compressed size:\t%.2f bits\n", tt->GetEncodedSize()); s = temp; LOG_MSG(s);
		delete tt;
	}
	// -- /End --

	// Cleanup
	delete ddb;
}

void TranslatorTH::ComputeTranslationTableEncodedSize() {
	char temp[500]; string s;

	// Load the data
	Database *dbL = Database::RetrieveDatabase(mConfig->Read<string>("dbLeft"), Uint16ItemSetType);
	Database *dbR = Database::RetrieveDatabase(mConfig->Read<string>("dbRight"), Uint16ItemSetType);
	DoubleDatabase *ddb = new DoubleDatabase(dbL, dbR); // takes ownership of dbL and dbR

	// Not very nice, possibly loading the data twice, but this is an easy solution for now; should give significant speedups when possible..
	if(ddb->GetAlphabetSize() < 129) {
		delete ddb;
		dbL = Database::RetrieveDatabase(mConfig->Read<string>("dbLeft"), BM128ItemSetType);
		dbR = Database::RetrieveDatabase(mConfig->Read<string>("dbRight"), BM128ItemSetType);
		ddb = new DoubleDatabase(dbL, dbR);
	}

	// Load TT from disk and compute encoded sizes
	TranslationTable *tt = TranslationTable::LoadTranslationTable(mConfig->Read<string>("ttDir") + mConfig->Read<string>("ttName") + ".tt", ddb);
	tt->Translate();

	// Output results to screen
	sprintf_s(temp, 500, " # rules:\t\t%u", tt->GetNumRules()); s = temp; LOG_MSG(s);
	sprintf_s(temp, 500, " Compressed size:\t%.2f bits", tt->GetEncodedSize()); s = temp; LOG_MSG(s);
	double stdSize = ddb->GetLeftDatabase()->GetStdDbSize() + ddb->GetRightDatabase()->GetStdDbSize();
	sprintf_s(temp, 500, " Standard size:\t\t%.2f bits", stdSize); s = temp; LOG_MSG(s);
	sprintf_s(temp, 500, " Compression ratio:\t%.2f%%\n", 100.0 * (tt->GetEncodedSize() / stdSize)); s = temp; LOG_MSG(s);
	delete tt;

	// Cleanup
	delete ddb;
}
