#include "../global.h"

#include <logger/Log.h>

#include "TranslationRule.h"
TranslationRule::TranslationRule(ItemSet *setL, const RuleDirection dir, ItemSet *setR) {
	mLeftSet = setL;
	mDirection = dir;
	mRightSet = setR;

	ComputeEncodedSize();
}
TranslationRule::TranslationRule(ItemSet *setL, const RuleDirection dir, ItemSet *setR, double encodedSize) {
	mLeftSet = setL;
	mDirection = dir;
	mRightSet = setR;
	mEncodedSize = encodedSize;
}
TranslationRule::TranslationRule(const TranslationRule &tr) {
	mLeftSet = tr.mLeftSet->Clone();
	mDirection = tr.mDirection;
	mRightSet = tr.mRightSet->Clone();

	mEncodedSize = tr.mEncodedSize;
}
TranslationRule::~TranslationRule() {
	delete mLeftSet;
	delete mRightSet;
}
TranslationRule* TranslationRule::Clone() {
	return new TranslationRule(*this);
}
void TranslationRule::SetDirection(const RuleDirection dir) {
	 mDirection = dir;
	 ComputeEncodedSize();
}
void TranslationRule::ComputeEncodedSize() {
	mEncodedSize = mLeftSet->GetStandardLength() + mRightSet->GetStandardLength();
	mEncodedSize += (mDirection==RuleBoth ? 1.0 : 2.0);
}
string TranslationRule::ToString() {
	string s = mLeftSet->ToString(false, false);
	if(!mLeftSet->IsEmpty()) s += " ";
	s += mDirection==RuleBoth ? "<->" : (mDirection==RuleToRight ? "->" : "<-");
	if(!mRightSet->IsEmpty()) s += " ";
	s += mRightSet->ToString(false, false);
	return s;
}
